#

import requests as rq

import sys
args = sys.argv[1:]
url = args.pop(0)
ofp = url.split("?")[0].split("/")[-1].strip()
if ofp=="":
    ofp = "index.html"
if len(args)>0:
    ofp = args.pop(0)

pass

rp = rq.get(url)
if rp.status_code!=200:
    print(f"get url error: {rp.status_code}")
    exit(0)
with open(ofp, 'wb') as f:
    f.write(rp.content)

print(f"save to '{ofp}'")


#coding=utf-8
import sys,re
from .base import Base
class Fetch:
    """
    命令行参数读取
    ft = Fetch("id,name,kind".split(","), {"a":"age"})
        or
    ft = Fetch(*xf.loads("[[id,name,kind],{a:age}]"))

    ft("001 zero life -a12".split(" ")) = {'id': '001', 'name': 'zero', 'kind': 'life', 'age': '12'}

    但更简单的方法是:
    xf.args("{id:'001', name: zero, kind: life, age: 12}".split(" ")) = {'id': '001', 'name': 'zero', 'kind': 'life', 'age': 12}
    python buildz.xf {id:'001', name: zero, kind: life, age: 12}
    就是对引号不太适用
    """
    def __init__(self, args = [], maps ={}, marks = set()):
        args = [maps[k] if k in maps else k for k in args]
        self.args = args
        self.maps = maps
        self.marks = marks
    def last(self, key):
        while key in self.maps:
            key = self.maps[key]
            if type(key) in (list, tuple):
                key = key[0]
        return key
    def des(self):
        cmd = " ".join([self.last(k) for k in self.args])
        adds = []
        marks = set()
        lasts = {}
        for key, val in self.maps.items():
            for k in [key, val]:
                if k in marks:
                    continue
                marks.add(k)
                last = self.last(k)
                param = f" <{last}>"
                if k in self.marks:
                    param = f"(set {last} as true)"
                if len(k)==1:
                    tags = [f"-{k}{param}"]
                else:
                    tags = [f"--{k}{param}"]
                    if k not in self.marks:
                        tags.append(f"--{k}=<{last}>")
                if last not in lasts:
                    lasts[last] = []
                lasts[last]+=tags
        for last, tags in lasts.items():
            tags = " | ".join(tags)
            tags = f"[{tags}]"
            adds.append(tags)
            #s = f"  {last}: {' '.join(tags)}"
        adds = [cmd]+adds
        rs = " ".join(adds)
        return rs
    def check(self, args, ks):
        rst = []
        for k in ks:
            if k not in args:
                rst.append(k)
        return rst
    def __call__(self, argv = None):
        args, maps = fetch(argv, self.marks)
        return self.fetch(args, maps)
    def fetch(self, args, maps):
        rst = {}
        for i in range(len(args)):
            if i >= len(self.args):
                break
            key = self.args[i]
            rst[key] = args[i]
        for key in maps:
            rst[key] = maps[key]
        keys = list(rst.keys())
        while len(keys)>0:
            key = keys.pop(0)
            if key in self.maps:
                rkeys = self.maps[key]
                if type(rkeys) not in (list, tuple):
                    rkeys = [rkeys]
                for rkey in rkeys:
                    assert rkey!=key, f'error rkey==key: {key}'
                    val = rst[key]
                    if rkey not in rst:
                        rst[rkey] = val
                    else:
                        tmp = rst[rkey]
                        if type(tmp)!=list:
                            tmp = [tmp]
                            rst[rkey] = tmp
                        tmp.append(val)
                    keys.append(rkey)
                    keys = list(set(keys))
                del rst[key]
        return rst

pass
class FetchX(Fetch):
    def __init__(self, *a, **b):
        super().__init__(a,b)

pass

def build_pt(pt, fc):
    st = "^"
    ed = "$"
    if pt[0]!=st:
        pt = st+pt
    if pt[-1]!=ed:
        pt = pt+ed
    def wfc(val):
        if re.match(pt, val) is None:
            return None, 0
        val = fc(val)
        return val, 1
    return wfc

pass
class ValDeals(Base):
    def init(self):
        self.deals = []
    def add(self, fc):
        self.deals.append(fc)
    def call(self, val):
        if type(val)!=str:
            return val
        for deal in self.deals:
            rst, done = deal(val)
            if done:
                return rst
        return val

pass
to_val = ValDeals()
to_val.add(build_pt("[\+\-]?\d+", int))
to_val.add(build_pt("[\+\-]?\d*\.\d+", float))
to_val.add(build_pt("[\+\-]?\d*(?:\.\d+)?e[\+\-]?\d+", float))
to_val.add(build_pt("null", lambda x:None))
to_val.add(build_pt("true", lambda x:True))
to_val.add(build_pt("false", lambda x:False))

def fetch(argv = None, marks = set()):
    r"""
    format: a b c -a 123 -b456 --c=789 +d  -x"??? ???" y z
    return: [a,b,c,y,z], {a:123,b:456,c:789,d:1,x:'??? ???'}
    """
    if argv is None:
        argv = sys.argv[1:]
    marks = set(marks)
    lists, maps = [],{}
    argv = [str(k).strip() for k in argv]
    argv = [k for k in argv if k!=""]
    i = 0
    def add_maps(maps, k, v):
        v = to_val(v)
        if k not in maps:
            maps[k] = []
        maps[k].append(v)
    while i<len(argv):
        v = argv[i]
        i+=1
        if v[0]!='-':
            v = to_val(v)
            lists.append(v)
            continue
        if v[1]=='-':
            v = v[2:]
            index = v.find("=")
            if index>=0:
                k = v[:index]
                v = v[index+1:]
            else:
                k = v
                v = None
            if k in marks:
                v = 1
            if v is None:
                assert i<len(argv), f"param '{k}' required value (-{k} val)"
                v = argv[i]
                i+=1
            add_maps(maps, k, v)
        else:
            ks = v[1:].strip()
            for k in ks:
                if k in marks:
                    v = 1
                else:
                    assert i<len(argv), f"param '{k}' required value (--{k} val | --{k}=val)"
                    v = argv[i]
                    i+=1
                add_maps(maps, k, v)
    for k in maps:
        v = maps[k]
        if len(v)==1:
            maps[k] = v[0]
    return lists, maps

pass





from buildz.gpuz.torch import DictCache
import torch, time
from torch import nn,optim
from torch.utils.data import DataLoader, Dataset
class TestDataset(Dataset):
    def __init__(self, num, dims, dtype=torch.float32):
        self.num = num
        self.dims = dims
        self.datas = torch.rand(num, dims).to(dtype)
        self.targets = torch.rand(num, dims).to(dtype)
        print(f"dataset:", self.datas.size(), self.datas.dtype)
    def __len__(self):
        return self.num
    def __getitem__(self, i):
        return self.datas[i], self.targets[i]
def datas(num, dims, batch_unit_size=10, dtype=torch.float32):
    dataset = TestDataset(num, dims, dtype)
    dataloader = DataLoader(dataset, batch_unit_size)
    return dataloader
def model(num_nets, dims, dtype=torch.float32):
    nets = [nn.Linear(dims,dims) for i in range(num_nets)]
    nets = nn.Sequential(*nets)
    nets=nets.to(dtype)
    return nets

loss_fn = torch.nn.MSELoss()
def train(data, md, num=10):
    md = md.cuda()
    dt, tdt = data
    dt = dt.cuda()
    tdt = tdt.cuda()
    opt = optim.Adam(md.parameters(), lr=0.001)
    md.train()
    curr=time.time()
    _loss=0
    for i in range(num):
        opt.zero_grad()
        out = md(dt)
        loss = loss_fn(out, tdt)
        loss.backward()
        _loss+=loss.cpu().item()
        opt.step()
    print(f"_loss: {_loss/num}")
    return (time.time()-curr)/num

def io(data, md, num=10):
    dt, tdt = data
    md.eval()
    curr=time.time()
    for i in range(num):
        dt=dt.cuda()
        # tdt=tdt.cuda()
        dt=dt.cpu()
        # tdt=tdt.cpu()
    return (time.time()-curr)/num

def eval(data, md, num=10):
    dt, tdt = data
    print(f"dt: {dt.size()}")
    print(f"md: {len(md)}")
    # exit(0)
    dt = dt.cuda()
    md=md.cuda()
    tdt = tdt.cuda()
    md.eval()
    curr=time.time()
    with torch.no_grad():
        for i in range(num):
            dt=dt.cuda()
            out = md(dt).cpu()
            dt=out
    return (time.time()-curr)/num

class Cost:
    def __init__(self, num_dt, num_net, dims, dtype):
        self.dts = datas(num_dt, dims, num_dt, dtype)
        print(f"dts:", len(self.dts))
        self.dt = list(self.dts)[0]
        self.net = model(num_net, dims, dtype)
    def eval(self, num=10):
        return eval(self.dt, self.net, num)
    def train(self, num=10):
        return train(self.dt, self.net, num)
    def io(self, num=10):
        return io(self.dt, self.net, num)

def build(num_dt, num_net, dims, dtype):
    return Cost(num_dt, num_net, dims, dtype)

def fmt(sz):
    units = "B,KB,MB,GB,TB".split(",")
    cnt=0
    while sz>1024:
        sz=sz/1024
        cnt+=1
    sz = "%.4f"%(sz,)
    s = f"{sz}{units[cnt]}"
    return s
def test():
    import sys
    args = sys.argv[1:]
    od = args.pop(0)[0].lower()
    num_dt = 10 if len(args)==0 else int(args.pop(0))
    num_net = 10 if len(args)==0 else int(args.pop(0))
    dims = 10 if len(args)==0 else int(args.pop(0))
    count = 10 if len(args)==0 else int(args.pop(0))
    dtypes = {}
    dtypes[16] = torch.float16
    dtypes[32] = torch.float32
    dtype = torch.float32 if len(args)==0 else dtypes[int(args.pop(0))]
    print(f"num_dt: {num_dt}, num_net: {num_net}, dims: {dims}, count: {count}, dtypes: {dtype}")
    count_dt = num_dt*dims
    count_net = num_net*dims*dims
    unit = 4 if dtype==torch.float32 else 2
    sz_dt = count_dt*unit
    sz_net = count_net*unit
    print(f"data size: {fmt(sz_dt)}, model size: {fmt(sz_net)}")
    cst = Cost(num_dt, num_net, dims, dtype)
    fcs = {}
    fcs['t'] = cst.train
    fcs['e'] = cst.eval
    fcs['i'] = cst.io
    print(f"start")
    sec = fcs[od](count)
    sec = "%.4f"%(sec,)
    print(f"data size: {fmt(sz_dt)}, model size: {fmt(sz_net)}")
    print(f"time cost: {sec} sec")

if __name__=="__main__":
    test()

pass


"""
python -m buildz.gpuz.test.cost i 4096 4096 1024 10 32

cst=Cost(4096, 4096, 1024, torch.float32)
"""